package ru.dbotthepony.kstarbound.client

import org.apache.logging.log4j.LogManager
import org.lwjgl.glfw.Callbacks
import org.lwjgl.glfw.GLFW
import org.lwjgl.glfw.GLFWErrorCallback
import org.lwjgl.opengl.GL46.*
import org.lwjgl.system.MemoryStack
import org.lwjgl.system.MemoryUtil
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.math.Matrix4f
import ru.dbotthepony.kstarbound.client.render.Camera
import ru.dbotthepony.kstarbound.client.render.TextAlignX
import ru.dbotthepony.kstarbound.client.render.TextAlignY
import ru.dbotthepony.kstarbound.math.Vector2f
import ru.dbotthepony.kstarbound.util.Color
import ru.dbotthepony.kstarbound.util.formatBytesShort
import kotlin.math.cos
import kotlin.math.sin

class StarboundClient : AutoCloseable {
	val window: Long
	val camera = Camera()
	var world: ClientWorld? = ClientWorld(this, 0L)

	var gameTerminated = false
		private set

	var viewportWidth = 800
		private set

	var viewportHeight = 600
		private set

	var viewportMatrixGUI = updateViewportMatrixA()
		private set

	var viewportMatrixGame = updateViewportMatrixB()
		private set

	private val startupTextList = ArrayList<String>()
	private var finishStartupRendering = System.currentTimeMillis() + 4000L

	fun pushLoadingText(text: String, replace: Boolean = false) {
		if (replace) {
			if (startupTextList.isEmpty()) {
				startupTextList.add(text)
			} else {
				startupTextList[startupTextList.size - 1] = text
			}
		} else {
			startupTextList.add(text)
		}

		finishStartupRendering = System.currentTimeMillis() + 4000L
	}

	private fun updateViewportMatrixA(): Matrix4f {
		return Matrix4f.ortho(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 0.1f, 100f)
	}

	private fun updateViewportMatrixB(): Matrix4f {
		return Matrix4f.orthoDirect(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 1f, 100f)
	}

	init {
		GLFWErrorCallback.create { error, description ->
			LOGGER.error("LWJGL error {}: {}", error, description)
		}.set()

		check(GLFW.glfwInit()) { "Unable to initialize GLFW" }

		GLFW.glfwDefaultWindowHints()

		GLFW.glfwWindowHint(GLFW.GLFW_VISIBLE, GLFW.GLFW_FALSE)
		GLFW.glfwWindowHint(GLFW.GLFW_RESIZABLE, GLFW.GLFW_TRUE)
		GLFW.glfwWindowHint(GLFW.GLFW_CONTEXT_VERSION_MAJOR, 4)
		GLFW.glfwWindowHint(GLFW.GLFW_CONTEXT_VERSION_MINOR, 6)
		GLFW.glfwWindowHint(GLFW.GLFW_OPENGL_PROFILE, GLFW.GLFW_OPENGL_CORE_PROFILE)

		window = GLFW.glfwCreateWindow(viewportWidth, viewportHeight, "KStarbound", MemoryUtil.NULL, MemoryUtil.NULL)
		require(window != MemoryUtil.NULL) { "Unable to create GLFW window" }
		startupTextList.add("Created GLFW window")

		GLFW.glfwSetKeyCallback(window) { window, key, scancode, action, mods ->
			if (key == GLFW.GLFW_KEY_ESCAPE || key == GLFW.GLFW_RELEASE) {
				GLFW.glfwSetWindowShouldClose(window, true)
			} else {
				camera.userInput(key, scancode, action, mods)
			}
		}

		GLFW.glfwSetFramebufferSizeCallback(window) { _, w, h ->
			viewportWidth = w
			viewportHeight = h
			viewportMatrixGUI = updateViewportMatrixA()
			viewportMatrixGame = updateViewportMatrixB()
			glViewport(0, 0, w, h)
		}

		val stack = MemoryStack.stackPush()

		try {
			val pWidth = stack.mallocInt(1)
			val pHeight = stack.mallocInt(1)

			GLFW.glfwGetWindowSize(window, pWidth, pHeight)

			val vidmode = GLFW.glfwGetVideoMode(GLFW.glfwGetPrimaryMonitor())!!

			GLFW.glfwSetWindowPos(
				window,
				(vidmode.width() - pWidth[0]) / 2,
				(vidmode.height() - pHeight[0]) / 2
			)
		} finally {
			stack.close()
		}

		GLFW.glfwMakeContextCurrent(window)

		// vsync
		GLFW.glfwSwapInterval(1)

		GLFW.glfwShowWindow(window)
		pushLoadingText("Initialized GLFW window")
	}

	val gl = GLStateTracker()

	fun ensureSameThread() = gl.ensureSameThread()

	init {
		pushLoadingText("Initialized OpenGL context")
		gl.clearColor = Color.SLATE_GREY

		gl.blend = true
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
	}

	var frameRenderTime = 1.0
		private set

	val framesPerSecond get() = 1.0 / frameRenderTime

	private val frameRenderTimes = DoubleArray(60) { 1.0 }
	private var frameRenderIndex = 0

	val averageFramesPerSecond: Double get() {
		var sum = 0.0

		for (value in frameRenderTimes) {
			sum += value
		}

		return frameRenderTimes.size / sum
	}

	val settings = ClientSettings()

	fun renderFrame(): Boolean {
		ensureSameThread()

		if (GLFW.glfwWindowShouldClose(window)) {
			close()
			return false
		}

		val measure = GLFW.glfwGetTime()

		glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT)
		gl.matrixStack.clear(viewportMatrixGame.toMutableMatrix())

		val mins = Vector2f((-viewportWidth / 2f) / settings.scale, (-viewportHeight / 2f) / settings.scale)
		val maxs = -mins

		gl.matrixStack.push()
			.translateWithScale(viewportWidth / 2f - camera.pos.x, viewportHeight / 2f - camera.pos.y) // центр экрана + координаты отрисовки мира
			.scale(x = settings.scale, y = settings.scale) // масштабируем до нужного размера

		world?.render(Vector2f.ZERO, mins = mins, maxs = maxs)

		gl.matrixStack.pop()

		gl.matrixStack.clear(viewportMatrixGUI.toMutableMatrix().translate(z = 2f))

		val thisTime = System.currentTimeMillis()

		if (startupTextList.isNotEmpty() && thisTime <= finishStartupRendering) {
			var alpha = 1f

			if (finishStartupRendering - thisTime < 1000L) {
				alpha = (finishStartupRendering - thisTime) / 1000f
			}

			gl.matrixStack.push()
			gl.matrixStack.translateWithScale(y = viewportHeight.toFloat())
			var shade = 255

			for (i in startupTextList.size - 1 downTo 0) {
				val size = gl.font.render(startupTextList[i], alignY = TextAlignY.BOTTOM, scale = 0.4f, color = Color.SHADES_OF_GRAY[shade].copy(alpha = alpha))
				gl.matrixStack.translateWithScale(y = -size.height * 1.2f)

				if (shade > 120) {
					shade -= 10
				}
			}

			gl.matrixStack.pop()
		}

		val runtime = Runtime.getRuntime()

		gl.font.render("FPS: %.2f".format(averageFramesPerSecond), scale = 0.4f)
		gl.font.render("Mem: ${formatBytesShort(runtime.totalMemory() - runtime.freeMemory())}", x = viewportWidth.toFloat(), scale = 0.4f, alignX = TextAlignX.RIGHT)

		GLFW.glfwSwapBuffers(window)
		GLFW.glfwPollEvents()

		camera.tick(GLFW.glfwGetTime() - measure)

		gl.cleanup()

		frameRenderTime = GLFW.glfwGetTime() - measure
		frameRenderTimes[++frameRenderIndex % frameRenderTimes.size] = frameRenderTime

		return true
	}

	private val terminateCallbacks = ArrayList<() -> Unit>()

	fun onTermination(lambda: () -> Unit) {
		terminateCallbacks.add(lambda)
	}

	override fun close() {
		if (gameTerminated)
			return

		if (window != MemoryUtil.NULL) {
			Callbacks.glfwFreeCallbacks(window)
			GLFW.glfwDestroyWindow(window)
		}

		GLFW.glfwTerminate()
		GLFW.glfwSetErrorCallback(null)?.free()

		gameTerminated = true

		for (callback in terminateCallbacks) {
			callback.invoke()
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger(StarboundClient::class.java)
	}
}
