package ru.dbotthepony.kstarbound.client.render

import ru.dbotthepony.kstarbound.PIXELS_IN_STARBOUND_UNITf
import ru.dbotthepony.kstarbound.client.ClientWorld
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.math.FloatMatrix
import ru.dbotthepony.kstarbound.math.Matrix4f
import ru.dbotthepony.kstarbound.math.Matrix4fStack
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ITileChunk
import kotlin.collections.ArrayList

/**
 * Псевдо zPos у фоновых тайлов
 *
 * Добавление этого числа к zPos гарантирует, что фоновые тайлы будут отрисованы
 * первыми (на самом дальнем плане)
 */
const val Z_LEVEL_BACKGROUND = 60000

class ChunkRenderer(val state: GLStateTracker, val chunk: Chunk, val world: ClientWorld? = null) : AutoCloseable, ILayeredRenderer {
	private inner class TileLayerRenderer(private val layerChangeset: () -> Int, private val isBackground: Boolean) : AutoCloseable {
		private val layers = TileLayerList()
		val bakedMeshes = ArrayList<Pair<BakedStaticMesh, Int>>()
		private var changeset = -1

		fun tesselateStatic(view: ITileChunk) {
			if (state.isSameThread()) {
				for (mesh in bakedMeshes) {
					mesh.first.close()
				}

				bakedMeshes.clear()
			} else {
				for (mesh in bakedMeshes) {
					unloadableBakedMeshes.add(mesh.first)
				}

				bakedMeshes.clear()
			}

			layers.clear()

			for ((pos, tile) in view.posToTile) {
				if (tile != null) {
					val renderer = state.tileRenderers.get(tile.def.materialName)
					renderer.tesselate(view, layers, pos, background = isBackground)
				}
			}
		}

		fun loadRenderers(view: ITileChunk) {
			for ((_, tile) in view.posToTile) {
				if (tile != null) {
					state.tileRenderers.get(tile.def.materialName)
				}
			}
		}

		fun uploadStatic(clear: Boolean = true) {
			for ((baked, builder, zLevel) in layers.buildList()) {
				bakedMeshes.add(BakedStaticMesh(baked, builder) to zLevel)
			}

			if (clear) {
				layers.clear()
			}
		}

		fun render(transform: FloatMatrix<*>) {
			for (mesh in bakedMeshes) {
				mesh.first.render(transform)
			}
		}

		fun bakeAndRender(transform: FloatMatrix<*>, provider: () -> ITileChunk) {
			if (changeset != layerChangeset.invoke()) {
				this.tesselateStatic(provider.invoke())
				this.uploadStatic()

				changeset = layerChangeset.invoke()
			}

			render(transform)
		}

		fun autoBake(provider: () -> ITileChunk) {
			if (changeset != layerChangeset.invoke()) {
				this.tesselateStatic(provider.invoke())
				this.uploadStatic()

				changeset = layerChangeset.invoke()
			}
		}

		fun autoUpload() {
			if (layers.isNotEmpty) {
				for (mesh in bakedMeshes) {
					mesh.first.close()
				}

				bakedMeshes.clear()

				for ((baked, builder, zLevel) in layers.buildList()) {
					bakedMeshes.add(BakedStaticMesh(baked, builder) to zLevel)
				}

				layers.clear()
			}
		}

		override fun close() {
			for (mesh in bakedMeshes) {
				mesh.first.close()
			}
		}
	}

	val transform = Matrix4f().translate(x = chunk.pos.x * CHUNK_SIZE * PIXELS_IN_STARBOUND_UNITf, y = chunk.pos.x * CHUNK_SIZE * PIXELS_IN_STARBOUND_UNITf)

	private val unloadableBakedMeshes = ArrayList<BakedStaticMesh>()

	private val foreground = TileLayerRenderer(chunk.foreground::changeset, isBackground = false)
	private val background = TileLayerRenderer(chunk.background::changeset, isBackground = true)

	private fun getForeground(): ITileChunk {
		return world?.getForegroundView(chunk.pos) ?: chunk.foreground
	}

	private fun getBackground(): ITileChunk {
		return world?.getBackgroundView(chunk.pos) ?: chunk.background
	}

	/**
	 * Тесселирует "статичную" геометрию в builders (к примеру тайлы).
	 *
	 * Может быть вызван вне рендер потока (ибо в любом случае он требует некой "стаитичности" данных в чанке)
	 * но только если до этого был вызыван loadRenderers() и геометрия чанка не поменялась
	 */
	fun tesselateStatic() {
		foreground.tesselateStatic(getForeground())
		background.tesselateStatic(getBackground())
	}

	/**
	 * Принудительно подгружает в GLStateTracker все необходимые рендереры (ибо им нужны текстуры и прочее)
	 *
	 * Вызывается перед tesselateStatic()
	 */
	fun loadRenderers() {
		unloadUnused()

		foreground.loadRenderers(getForeground())
		background.loadRenderers(getBackground())
	}

	private fun unloadUnused() {
		if (unloadableBakedMeshes.size != 0) {
			for (baked in unloadableBakedMeshes) {
				baked.close()
			}

			unloadableBakedMeshes.clear()
		}
	}

	fun uploadStatic(clear: Boolean = true) {
		unloadUnused()

		foreground.uploadStatic(clear)
		background.uploadStatic(clear)
	}

	/**
	 * Отрисовывает всю геометрию напрямую
	 */
	fun render(transform: FloatMatrix<*> = state.matrixStack.last) {
		unloadUnused()

		background.render(transform)
		foreground.render(transform)
	}

	/**
	 * Отрисовывает всю геометрию напрямую, с проверкой, изменился ли чанк
	 */
	fun bakeAndRender(transform: FloatMatrix<*> = state.matrixStack.last) {
		unloadUnused()

		background.bakeAndRender(transform, this::getBackground)
		foreground.bakeAndRender(transform, this::getForeground)
	}

	/**
	 * Запекает всю геометрию напрямую, с проверкой, изменился ли чанк,
	 * и загружает её, если вызвано в рендер потоке
	 */
	fun autoBakeStatic() {
		if (state.isSameThread())
			unloadUnused()

		background.autoBake(this::getBackground)
		foreground.autoBake(this::getForeground)

		if (state.isSameThread())
			autoUploadStatic()
	}

	/**
	 * Загружает в видеопамять всю геометрию напрямую, если есть что загружать
	 */
	fun autoUploadStatic() {
		unloadUnused()

		background.autoUpload()
		foreground.autoUpload()
	}

	private val meshDeque = ArrayDeque<Pair<BakedStaticMesh, Int>>()

	override fun renderLayerFromStack(zPos: Int, transform: Matrix4fStack): Int {
		if (meshDeque.isEmpty())
			return -1

		transform.push().translateWithScale(x = chunk.pos.x * CHUNK_SIZE * PIXELS_IN_STARBOUND_UNITf, y = chunk.pos.y * CHUNK_SIZE * PIXELS_IN_STARBOUND_UNITf)
		var pair = meshDeque.last()

		while (pair.second >= zPos) {
			pair.first.render(transform.last)

			meshDeque.removeLast()

			if (meshDeque.isEmpty()) {
				transform.pop()
				return -1
			}

			pair = meshDeque.last()
		}

		transform.pop()
		return meshDeque.last().second
	}

	override fun bottomMostZLevel(): Int {
		if (meshDeque.isEmpty()) {
			return -1
		}

		return meshDeque.last().second
	}

	override fun prepareForLayeredRender() {
		meshDeque.clear()

		for ((baked, zLevel) in background.bakedMeshes) {
			meshDeque.add(baked to (zLevel + Z_LEVEL_BACKGROUND))
		}

		meshDeque.addAll(foreground.bakedMeshes)

		meshDeque.sortBy {
			return@sortBy it.second
		}
	}

	override fun close() {
		background.close()
		foreground.close()
	}
}
