package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.math.Vector2i

/**
 * Возвращает кортеж чанка, который содержит родителя (мир) и соседей (кортежи чанков)
 */
interface IWorldChunkTuple {
	val world: World<*>
	val chunk: Chunk
	val top: IWorldChunkTuple?
	val left: IWorldChunkTuple?
	val right: IWorldChunkTuple?
	val bottom: IWorldChunkTuple?
}

interface IMutableWorldChunkTuple : IWorldChunkTuple {
	override var top: IWorldChunkTuple?
	override var left: IWorldChunkTuple?
	override var right: IWorldChunkTuple?
	override var bottom: IWorldChunkTuple?
}

data class WorldChunkTuple(
	override val world: World<*>,
	override val chunk: Chunk,
	override val top: IWorldChunkTuple?,
	override val left: IWorldChunkTuple?,
	override val right: IWorldChunkTuple?,
	override val bottom: IWorldChunkTuple?,
) : IWorldChunkTuple

open class MutableWorldChunkTuple(
	override val world: World<*>,
	override val chunk: Chunk,
	override var top: IWorldChunkTuple?,
	override var left: IWorldChunkTuple?,
	override var right: IWorldChunkTuple?,
	override var bottom: IWorldChunkTuple?,
) : IMutableWorldChunkTuple

@Suppress("WeakerAccess")
abstract class World<T : IMutableWorldChunkTuple>(val seed: Long = 0L) {
	protected val chunkMap = HashMap<ChunkPos, T>()
	protected var lastAccessedChunk: T? = null

	protected abstract fun tupleFactory(
		chunk: Chunk,
		top: IWorldChunkTuple?,
		left: IWorldChunkTuple?,
		right: IWorldChunkTuple?,
		bottom: IWorldChunkTuple?,
	): T

	protected fun getChunkInternal(pos: ChunkPos): T? {
		if (lastAccessedChunk?.chunk?.pos == pos) {
			return lastAccessedChunk
		}

		return chunkMap[pos]
	}

	open fun getChunk(pos: ChunkPos): IWorldChunkTuple? {
		val getTuple = getChunkInternal(pos)

		if (getTuple != null)
			return WorldChunkTuple(
				world = getTuple.world,
				chunk = getTuple.chunk,
				top = getTuple.top,
				left = getTuple.left,
				right = getTuple.right,
				bottom = getTuple.bottom,
			)

		return null
	}

	protected open fun computeIfAbsentInternal(pos: ChunkPos): T {
		if (lastAccessedChunk?.chunk?.pos == pos) {
			return lastAccessedChunk!!
		}

		return chunkMap.computeIfAbsent(pos) lazy@{
			val chunk = Chunk(this, pos)

			val top = getChunkInternal(pos.up())
			val left = getChunkInternal(pos.left())
			val right = getChunkInternal(pos.right())
			val bottom = getChunkInternal(pos.down())

			val tuple = tupleFactory(
				chunk = chunk,
				top = top,
				left = left,
				right = right,
				bottom = bottom,
			)

			top?.bottom = tuple
			left?.right = tuple
			right?.left = tuple
			bottom?.top = tuple

			lastAccessedChunk = tuple

			return@lazy tuple
		}
	}

	open fun computeIfAbsent(pos: ChunkPos): IWorldChunkTuple {
		val getTuple = computeIfAbsentInternal(pos)

		return WorldChunkTuple(
			world = getTuple.world,
			chunk = getTuple.chunk,
			top = getTuple.top,
			left = getTuple.left,
			right = getTuple.right,
			bottom = getTuple.bottom,
		)
	}

	open fun getForegroundView(pos: ChunkPos): TileChunkView? {
		val get = getChunkInternal(pos) ?: return null

		return TileChunkView(
			center = 			get.chunk.foreground,
			left = 				get.left?.chunk?.foreground,
			top = 				get.top?.chunk?.foreground,
			topLeft = 			getChunkInternal(pos.up().left())?.chunk?.foreground,
			topRight = 			getChunkInternal(pos.up().right())?.chunk?.foreground,
			right = 			get.right?.chunk?.foreground,
			bottom = 			get.bottom?.chunk?.foreground,
			bottomLeft = 		getChunkInternal(pos.down().left())?.chunk?.foreground,
			bottomRight = 		getChunkInternal(pos.down().right())?.chunk?.foreground,
		)
	}

	open fun getBackgroundView(pos: ChunkPos): TileChunkView? {
		val get = getChunkInternal(pos) ?: return null

		return TileChunkView(
			center = 			get.chunk.background,
			left = 				get.left?.chunk?.background,
			top = 				get.top?.chunk?.background,
			topLeft = 			getChunkInternal(pos.up().left())?.chunk?.background,
			topRight = 			getChunkInternal(pos.up().right())?.chunk?.background,
			right = 			get.right?.chunk?.background,
			bottom = 			get.bottom?.chunk?.background,
			bottomLeft = 		getChunkInternal(pos.down().left())?.chunk?.background,
			bottomRight = 		getChunkInternal(pos.down().right())?.chunk?.background,
		)
	}

	fun getTile(pos: Vector2i): ChunkTile? {
		return getChunkInternal(ChunkPos.fromTilePosition(pos))?.chunk?.foreground?.get(pos.x, pos.y)
	}

	fun setTile(pos: Vector2i, tile: TileDefinition?): IWorldChunkTuple {
		val chunk = computeIfAbsentInternal(ChunkPos.fromTilePosition(pos))
		chunk.chunk.foreground[pos.x and CHUNK_SIZE_FF, pos.y and CHUNK_SIZE_FF] = tile
		return chunk
	}

	fun getBackgroundTile(pos: Vector2i): ChunkTile? {
		return getChunkInternal(ChunkPos.fromTilePosition(pos))?.chunk?.background?.get(pos.x, pos.y)
	}

	fun setBackgroundTile(pos: Vector2i, tile: TileDefinition?): IWorldChunkTuple {
		val chunk = computeIfAbsentInternal(ChunkPos.fromTilePosition(pos))
		chunk.chunk.background[pos.x and CHUNK_SIZE_FF, pos.y and CHUNK_SIZE_FF] = tile
		return chunk
	}
}
